#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
4200 - Module for testing scrollable cursors
"""

import unittest

import oracledb
import test_env

@unittest.skipIf(test_env.get_is_thin(),
                 "thin mode doesn't support scrollable cursors yet")
class TestCase(test_env.BaseTestCase):

    def test_4200_create_scrollable_cursor(self):
        "4200 - test creating a scrollable cursor"
        cursor = self.connection.cursor()
        self.assertEqual(cursor.scrollable, False)
        cursor = self.connection.cursor(True)
        self.assertEqual(cursor.scrollable, True)
        cursor = self.connection.cursor(scrollable=True)
        self.assertEqual(cursor.scrollable, True)
        cursor.scrollable = False
        self.assertEqual(cursor.scrollable, False)

    def test_4201_scroll_absolute_exception_after(self):
        "4201 - test scrolling absolute yields an exception (after result set)"
        cursor = self.connection.cursor(scrollable=True)
        cursor.arraysize = self.cursor.arraysize
        cursor.execute("""
                select NumberCol
                from TestNumbers
                order by IntCol""")
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPI-1027:",
                               cursor.scroll, 12, "absolute")

    def test_4202_scroll_absolute_in_buffer(self):
        "4202 - test scrolling absolute (when in buffers)"
        cursor = self.connection.cursor(scrollable=True)
        cursor.arraysize = self.cursor.arraysize
        cursor.execute("""
                select NumberCol
                from TestNumbers
                order by IntCol""")
        cursor.fetchmany()
        self.assertTrue(cursor.arraysize > 1,
                "array size must exceed 1 for this test to work correctly")
        cursor.scroll(1, mode="absolute")
        row = cursor.fetchone()
        self.assertEqual(row[0], 1.25)
        self.assertEqual(cursor.rowcount, 1)

    def test_4203_scroll_absolute_not_in_buffer(self):
        "4203 - test scrolling absolute (when not in buffers)"
        cursor = self.connection.cursor(scrollable=True)
        cursor.arraysize = self.cursor.arraysize
        cursor.execute("""
                select NumberCol
                from TestNumbers
                order by IntCol""")
        cursor.scroll(6, mode="absolute")
        row = cursor.fetchone()
        self.assertEqual(row[0], 7.5)
        self.assertEqual(cursor.rowcount, 6)

    def test_4204_scroll_first_in_buffer(self):
        "4204 - test scrolling to first row in result set (in buffers)"
        cursor = self.connection.cursor(scrollable=True)
        cursor.arraysize = self.cursor.arraysize
        cursor.execute("""
                select NumberCol
                from TestNumbers
                order by IntCol""")
        cursor.fetchmany()
        cursor.scroll(mode="first")
        row = cursor.fetchone()
        self.assertEqual(row[0], 1.25)
        self.assertEqual(cursor.rowcount, 1)

    def test_4205_scroll_first_not_in_buffer(self):
        "4205 - test scrolling to first row in result set (not in buffers)"
        cursor = self.connection.cursor(scrollable=True)
        cursor.arraysize = self.cursor.arraysize
        cursor.execute("""
                select NumberCol
                from TestNumbers
                order by IntCol""")
        cursor.fetchmany()
        cursor.fetchmany()
        cursor.scroll(mode="first")
        row = cursor.fetchone()
        self.assertEqual(row[0], 1.25)
        self.assertEqual(cursor.rowcount, 1)

    def test_4206_scroll_last(self):
        "4206 - test scrolling to last row in result set"
        cursor = self.connection.cursor(scrollable=True)
        cursor.arraysize = self.cursor.arraysize
        cursor.execute("""
                select NumberCol
                from TestNumbers
                order by IntCol""")
        cursor.scroll(mode="last")
        row = cursor.fetchone()
        self.assertEqual(row[0], 12.5)
        self.assertEqual(cursor.rowcount, 10)

    def test_4207_scroll_relative_exception_after(self):
        "4207 - test scrolling relative yields an exception (after result set)"
        cursor = self.connection.cursor(scrollable=True)
        cursor.arraysize = self.cursor.arraysize
        cursor.execute("""
                select NumberCol
                from TestNumbers
                order by IntCol""")
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPI-1027:",
                               cursor.scroll, 15)

    def test_4208_scroll_relative_exception_before(self):
        "4208 - test scrolling relative yields exception (before result set)"
        cursor = self.connection.cursor(scrollable=True)
        cursor.arraysize = self.cursor.arraysize
        cursor.execute("""
                select NumberCol
                from TestNumbers
                order by IntCol""")
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPI-1027:",
                               cursor.scroll, -5)

    def test_4209_scroll_relative_in_buffer(self):
        "4209 - test scrolling relative (when in buffers)"
        cursor = self.connection.cursor(scrollable=True)
        cursor.arraysize = self.cursor.arraysize
        cursor.execute("""
                select NumberCol
                from TestNumbers
                order by IntCol""")
        cursor.fetchmany()
        message = "array size must exceed 1 for this test to work correctly"
        self.assertTrue(cursor.arraysize > 1, message)
        cursor.scroll(2 - cursor.rowcount)
        row = cursor.fetchone()
        self.assertEqual(row[0], 2.5)
        self.assertEqual(cursor.rowcount, 2)

    def test_4210_scroll_relative_not_in_buffer(self):
        "4210 - test scrolling relative (when not in buffers)"
        cursor = self.connection.cursor(scrollable=True)
        cursor.arraysize = self.cursor.arraysize
        cursor.execute("""
                select NumberCol
                from TestNumbers
                order by IntCol""")
        cursor.fetchmany()
        cursor.fetchmany()
        message = "array size must exceed 1 for this test to work correctly"
        self.assertTrue(cursor.arraysize > 1, message)
        cursor.scroll(3 - cursor.rowcount)
        row = cursor.fetchone()
        self.assertEqual(row[0], 3.75)
        self.assertEqual(cursor.rowcount, 3)

    def test_4211_scroll_no_rows(self):
        "4211 - test scrolling when there are no rows"
        self.cursor.execute("truncate table TestTempTable")
        cursor = self.connection.cursor(scrollable=True)
        cursor.execute("select * from TestTempTable")
        cursor.scroll(mode="last")
        self.assertEqual(cursor.fetchall(), [])
        cursor.scroll(mode="first")
        self.assertEqual(cursor.fetchall(), [])
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPI-1027:",
                               cursor.scroll, 1, mode="absolute")

    def test_4212_scroll_differing_array_and_fetch_sizes(self):
        "4212 - test scrolling with differing array and fetch array sizes"
        self.cursor.execute("truncate table TestTempTable")
        for i in range(30):
            self.cursor.execute("""
                    insert into TestTempTable (IntCol, StringCol1)
                    values (:1, null)""",
                    (i + 1,))
        for arraysize in range(1, 6):
            cursor = self.connection.cursor(scrollable = True)
            cursor.arraysize = arraysize
            cursor.execute("select IntCol from TestTempTable order by IntCol")
            for num_rows in range(1, arraysize + 1):
                cursor.scroll(15, "absolute")
                rows = cursor.fetchmany(num_rows)
                self.assertEqual(rows[0][0], 15)
                self.assertEqual(cursor.rowcount, 15 + num_rows - 1)
                cursor.scroll(9)
                rows = cursor.fetchmany(num_rows)
                num_rows_fetched = len(rows)
                self.assertEqual(rows[0][0], 15 + num_rows + 8)
                self.assertEqual(cursor.rowcount,
                                 15 + num_rows + num_rows_fetched + 7)
                cursor.scroll(-12)
                rows = cursor.fetchmany(num_rows)
                count = 15 + num_rows + num_rows_fetched - 5
                self.assertEqual(rows[0][0], count)
                count = 15 + num_rows + num_rows_fetched + num_rows - 6
                self.assertEqual(cursor.rowcount, count)

    def test_4213_scroll_with_invalid_mode(self):
        "4213 - test calling scroll() with invalid mode"
        cursor = self.connection.cursor(scrollable=True)
        cursor.arraysize = self.cursor.arraysize
        cursor.execute("""
                select NumberCol
                from TestNumbers
                order by IntCol""")
        cursor.fetchmany()
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2009:",
                               cursor.scroll, mode="middle")

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2023, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# base_impl.pxd
#
# Cython definition file defining the base classes from which the thick and
# thin implementations derive their classes.
#------------------------------------------------------------------------------

# cython: language_level=3

from libc.stdint cimport int8_t, int16_t, int32_t, int64_t
from libc.stdint cimport uint8_t, uint16_t, uint32_t, uint64_t

cdef enum:
    NUM_TYPE_FLOAT = 0
    NUM_TYPE_INT = 1
    NUM_TYPE_DECIMAL = 2
    NUM_TYPE_STR = 3

cdef enum:
    DB_TYPE_NUM_BFILE = 2020
    DB_TYPE_NUM_BINARY_DOUBLE = 2008
    DB_TYPE_NUM_BINARY_FLOAT = 2007
    DB_TYPE_NUM_BINARY_INTEGER = 2009
    DB_TYPE_NUM_BLOB = 2019
    DB_TYPE_NUM_BOOLEAN = 2022
    DB_TYPE_NUM_CHAR = 2003
    DB_TYPE_NUM_CLOB = 2017
    DB_TYPE_NUM_CURSOR = 2021
    DB_TYPE_NUM_DATE = 2011
    DB_TYPE_NUM_INTERVAL_DS = 2015
    DB_TYPE_NUM_INTERVAL_YM = 2016
    DB_TYPE_NUM_JSON = 2027
    DB_TYPE_NUM_LONG_NVARCHAR = 2031
    DB_TYPE_NUM_LONG_RAW = 2025
    DB_TYPE_NUM_LONG_VARCHAR = 2024
    DB_TYPE_NUM_NCHAR = 2004
    DB_TYPE_NUM_NCLOB = 2018
    DB_TYPE_NUM_NUMBER = 2010
    DB_TYPE_NUM_NVARCHAR = 2002
    DB_TYPE_NUM_OBJECT = 2023
    DB_TYPE_NUM_RAW = 2006
    DB_TYPE_NUM_ROWID = 2005
    DB_TYPE_NUM_TIMESTAMP = 2012
    DB_TYPE_NUM_TIMESTAMP_LTZ = 2014
    DB_TYPE_NUM_TIMESTAMP_TZ = 2013
    DB_TYPE_NUM_UNKNOWN = 0
    DB_TYPE_NUM_UROWID = 2030
    DB_TYPE_NUM_VARCHAR = 2001


cdef class ApiType:
    cdef:
        readonly str name
        tuple dbtypes


cdef class DbType:
    cdef:
        readonly uint32_t num
        readonly str name
        readonly uint32_t default_size
        uint32_t _buffer_size_factor
        str _ora_name
        uint8_t _ora_type_num
        uint8_t _csfrm

    @staticmethod
    cdef DbType _from_num(uint32_t num)

    @staticmethod
    cdef DbType _from_ora_name(str name)

    @staticmethod
    cdef DbType _from_ora_type_and_csfrm(uint8_t ora_type_num, uint8_t csfrm)


cdef class Address:
    cdef:
        public str host
        public uint32_t port
        public str protocol
        public str https_proxy
        public uint32_t https_proxy_port

    cdef str build_connect_string(self)


cdef class AddressList:
    cdef:
        public list addresses
        bint source_route
        bint load_balance
        int lru_index

    cdef bint _uses_tcps(self)
    cdef str build_connect_string(self)


cdef class Description:
    cdef:
        public list address_lists
        public bint source_route
        public bint load_balance
        public int lru_index
        public uint32_t expire_time
        public uint32_t retry_count
        public uint32_t retry_delay
        public double tcp_connect_timeout
        public str service_name
        public str server_type
        public str sid
        public str cclass
        public uint32_t purity
        public bint ssl_server_dn_match
        public str ssl_server_cert_dn
        public str wallet_location

    cdef str _build_duration_str(self, double value)
    cdef str build_connect_string(self, str cid=*)


cdef class DescriptionList:
    cdef:
        public list descriptions
        bint source_route
        bint load_balance
        int lru_index

    cdef str build_connect_string(self)


cdef class TnsnamesFile:
    cdef:
        str file_name
        int mtime
        dict entries


cdef class ConnectParamsImpl:
    cdef:
        public str config_dir
        public str user
        public str proxy_user
        public bint events
        public bint externalauth
        public uint32_t mode
        public str edition
        public list appcontext
        public str tag
        public bint matchanytag
        public list shardingkey
        public list supershardingkey
        public uint32_t stmtcachesize
        public bint disable_oob
        public DescriptionList description_list
        uint64_t _external_handle
        public str debug_jdwp
        object access_token_callback
        object access_token_expires
        Description _default_description
        Address _default_address
        bytearray _password
        bytearray _password_obfuscator
        bytearray _new_password
        bytearray _new_password_obfuscator
        bytearray _wallet_password
        bytearray _wallet_password_obfuscator
        bytearray _token
        bytearray _token_obfuscator
        bytearray _private_key
        bytearray _private_key_obfuscator

    cdef int _check_credentials(self) except -1
    cdef int _copy(self, ConnectParamsImpl other_params) except -1
    cdef bytes _get_new_password(self)
    cdef bytearray _get_obfuscator(self, str secret_value)
    cdef bytes _get_password(self)
    cdef str _get_private_key(self)
    cdef TnsnamesFile _get_tnsnames_file(self)
    cdef str _get_token(self)
    cdef object _get_token_expires(self, str token)
    cdef str _get_wallet_password(self)
    cdef int _parse_connect_string(self, str connect_string) except -1
    cdef int _process_connect_descriptor(self, dict args) except -1
    cdef int _set_access_token(self, object val, int error_num) except -1
    cdef int _set_access_token_param(self, object val) except -1
    cdef int _set_new_password(self, str password) except -1
    cdef int _set_password(self, str password) except -1
    cdef int _set_wallet_password(self, str password) except -1
    cdef bytearray _xor_bytes(self, bytearray a, bytearray b)


cdef class PoolParamsImpl(ConnectParamsImpl):
    cdef:
        public uint32_t min
        public uint32_t max
        public uint32_t increment
        public type connectiontype
        public uint32_t getmode
        public bint homogeneous
        public uint32_t timeout
        public uint32_t wait_timeout
        public uint32_t max_lifetime_session
        public object session_callback
        public uint32_t max_sessions_per_shard
        public bint soda_metadata_cache
        public int ping_interval


cdef class BaseConnImpl:
    cdef:
        readonly str username
        readonly str dsn
        public object inputtypehandler
        public object outputtypehandler
        public bint autocommit
        public bint invoke_session_callback

    cdef object _check_value(self, DbType dbtype, BaseDbObjectTypeImpl objtype,
                             object value, bint* is_ok)


cdef class BasePoolImpl:
    cdef:
        readonly str dsn
        readonly bint homogeneous
        readonly uint32_t increment
        readonly uint32_t min
        readonly uint32_t max
        readonly str username
        readonly str name
        ConnectParamsImpl connect_params


cdef class BaseCursorImpl:
    cdef:
        readonly str statement
        readonly uint64_t rowcount
        public uint32_t arraysize
        public uint32_t prefetchrows
        public object inputtypehandler
        public object outputtypehandler
        public object rowfactory
        public bint scrollable
        public list fetch_vars
        public list fetch_var_impls
        public list bind_vars
        public type bind_style
        public dict bind_vars_by_name
        uint32_t _buffer_rowcount
        uint32_t _buffer_index
        uint32_t _fetch_array_size
        bint _more_rows_to_fetch

    cdef int _bind_values(self, object cursor, object type_handler,
                          object params, uint32_t num_rows, uint32_t row_num,
                          bint defer_type_assignment) except -1
    cdef int _bind_values_by_name(self, object cursor, object type_handler,
                                  dict params, uint32_t num_rows,
                                  uint32_t row_num,
                                  bint defer_type_assignment) except -1
    cdef int _bind_values_by_position(self, object cursor, object type_handler,
                                      object params, uint32_t num_rows,
                                      uint32_t row_num,
                                      bint defer_type_assignment) except -1
    cdef int _close(self, bint in_del) except -1
    cdef int _create_fetch_var(self, object conn, object cursor,
                               object type_handler, ssize_t pos,
                               FetchInfo fetch_info) except -1
    cdef object _create_row(self)
    cdef BaseVarImpl _create_var_impl(self, object conn)
    cdef int _fetch_rows(self, object cursor) except -1
    cdef BaseConnImpl _get_conn_impl(self)
    cdef object _get_input_type_handler(self)
    cdef object _get_output_type_handler(self)
    cdef int _init_fetch_vars(self, uint32_t num_columns) except -1
    cdef bint _is_plsql(self)
    cdef int _perform_binds(self, object conn, uint32_t num_execs) except -1
    cdef int _reset_bind_vars(self, uint32_t num_rows) except -1
    cdef int _verify_var(self, object var) except -1


cdef class FetchInfo:
    cdef:
        int16_t _precision
        int16_t _scale
        uint32_t _buffer_size
        uint32_t _size
        bint _nulls_allowed
        str _name
        DbType _dbtype
        BaseDbObjectTypeImpl _objtype


cdef class BaseVarImpl:
    cdef:
        readonly str name
        readonly int16_t precision
        readonly int16_t scale
        readonly uint32_t num_elements
        readonly object inconverter
        readonly object outconverter
        readonly uint32_t size
        readonly uint32_t buffer_size
        readonly bint bypass_decode
        readonly bint is_array
        readonly bint nulls_allowed
        public uint32_t num_elements_in_array
        readonly DbType dbtype
        readonly BaseDbObjectTypeImpl objtype
        BaseConnImpl _conn_impl
        int _preferred_num_type
        FetchInfo _fetch_info
        bint _is_value_set

    cdef int _bind(self, object conn, BaseCursorImpl cursor,
                   uint32_t num_execs, object name, uint32_t pos) except -1
    cdef int _check_and_set_scalar_value(self, uint32_t pos, object value,
                                         bint* was_set) except -1
    cdef int _check_and_set_value(self, uint32_t pos, object value,
                                  bint* was_set) except -1
    cdef int _finalize_init(self) except -1
    cdef list _get_array_value(self)
    cdef object _get_scalar_value(self, uint32_t pos)
    cdef int _on_reset_bind(self, uint32_t num_rows) except -1
    cdef int _resize(self, uint32_t new_size) except -1
    cdef int _set_scalar_value(self, uint32_t pos, object value) except -1
    cdef int _set_num_elements_in_array(self, uint32_t num_elements) except -1
    cdef int _set_type_info_from_type(self, object typ) except -1
    cdef int _set_type_info_from_value(self, object value,
                                       bint is_plsql) except -1


cdef class BaseLobImpl:
    cdef:
        readonly DbType dbtype


cdef class BaseDbObjectTypeImpl:
    cdef:
        readonly str schema
        readonly str name
        readonly str package_name
        readonly list attrs
        readonly bint is_collection
        readonly dict attrs_by_name
        readonly DbType element_dbtype
        readonly BaseDbObjectTypeImpl element_objtype
        readonly BaseConnImpl _conn_impl


cdef class BaseDbObjectAttrImpl:
    cdef:
        readonly str name
        readonly DbType dbtype
        readonly BaseDbObjectTypeImpl objtype


cdef class BaseDbObjectImpl:
    cdef:
        readonly BaseDbObjectTypeImpl type


cdef class BaseSodaDbImpl:
    cdef:
        object _conn


cdef class BaseSodaCollImpl:
    cdef:
        readonly str name


cdef class BaseSodaDocImpl:
    pass


cdef class BaseSodaDocCursorImpl:
    pass


cdef class BaseQueueImpl:
    cdef:
        readonly str name
        readonly BaseDbObjectTypeImpl payload_type
        readonly BaseDeqOptionsImpl deq_options_impl
        readonly BaseEnqOptionsImpl enq_options_impl
        readonly bint is_json


cdef class BaseDeqOptionsImpl:
    pass


cdef class BaseEnqOptionsImpl:
    pass


cdef class BaseMsgPropsImpl:
    cdef:
        public object payload


cdef class BaseSubscrImpl:
    cdef:
        readonly object callback
        readonly object connection
        readonly uint32_t namespace
        readonly str name
        readonly uint32_t protocol
        readonly str ip_address
        readonly uint32_t port
        readonly uint32_t timeout
        readonly uint32_t operations
        readonly uint32_t qos
        readonly uint64_t id
        readonly uint8_t grouping_class
        readonly uint32_t grouping_value
        readonly uint8_t grouping_type
        readonly bint client_initiated


cdef class BindVar:
    cdef:
        object var
        BaseVarImpl var_impl
        object name
        ssize_t pos
        bint has_value

    cdef int _create_var_from_type(self, object conn,
                                   BaseCursorImpl cursor_impl,
                                   object value) except -1
    cdef int _create_var_from_value(self, object conn,
                                    BaseCursorImpl cursor_impl, object value,
                                    uint32_t num_elements) except -1
    cdef int _set_by_type(self, object conn, BaseCursorImpl cursor_impl,
                          object typ) except -1
    cdef int _set_by_value(self, object conn, BaseCursorImpl cursor_impl,
                           object cursor, object value, object type_handler,
                           uint32_t row_num, uint32_t num_elements,
                           bint defer_type_assignment) except -1

cdef object get_exception_class(int32_t code)



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# thick_impl.pyx
#
# Cython file for interfacing with ODPI-C.
#------------------------------------------------------------------------------

# cython: language_level=3

cimport cython
cimport cpython
cimport cpython.datetime as cydatetime

import datetime
import decimal

cydatetime.import_datetime()

from . import constants, driver_mode, errors, exceptions
from .subscr import Message, MessageQuery, MessageRow, MessageTable
from .defaults import defaults

from . import __version__ as VERSION

from .base_impl cimport get_exception_class
from .base_impl cimport BaseConnImpl, BaseCursorImpl, BaseVarImpl, DbType
from .base_impl cimport BaseDbObjectTypeImpl, BaseDbObjectAttrImpl
from .base_impl cimport BaseDbObjectImpl, BaseLobImpl, BasePoolImpl
from .base_impl cimport BaseSodaDbImpl, BaseSodaCollImpl, BaseSodaDocImpl
from .base_impl cimport BaseSodaDocCursorImpl, BaseQueueImpl
from .base_impl cimport BaseDeqOptionsImpl, BaseEnqOptionsImpl
from .base_impl cimport BaseMsgPropsImpl, BaseSubscrImpl, BindVar, FetchInfo
from .base_impl cimport ConnectParamsImpl, PoolParamsImpl
from .base_impl cimport NUM_TYPE_FLOAT, NUM_TYPE_INT, NUM_TYPE_DECIMAL
from libc.string cimport memchr, memset

include "impl/thick/odpi.pxd"

cdef type PY_TYPE_DATE = datetime.date
cdef type PY_TYPE_DATETIME = datetime.datetime
cdef type PY_TYPE_DECIMAL = decimal.Decimal
cdef type PY_TYPE_DB_OBJECT
cdef type PY_TYPE_LOB
cdef type PY_TYPE_TIMEDELTA = datetime.timedelta

cdef dpiContext *driver_context = NULL

driver_context_params = None
client_version = None

include "impl/thick/buffer.pyx"
include "impl/thick/connection.pyx"
include "impl/thick/pool.pyx"
include "impl/thick/cursor.pyx"
include "impl/thick/lob.pyx"
include "impl/thick/json.pyx"
include "impl/thick/var.pyx"
include "impl/thick/dbobject.pyx"
include "impl/thick/soda.pyx"
include "impl/thick/queue.pyx"
include "impl/thick/subscr.pyx"
include "impl/thick/utils.pyx"



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# constructors.py
#
# Contains the constructors mandated by the Python Database API.
#------------------------------------------------------------------------------

import datetime

from . import errors

# synonyms for the types mandated by the database API
Binary = bytes
Date = datetime.date
Timestamp = datetime.datetime


def Time(hour: int, minute: int, second: int) -> None:
    """
    Constructor mandated by the database API for creating a time value. Since
    Oracle doesn't support time only values, an exception is raised when this
    method is called.
    """
    errors._raise_err(errors.ERR_TIME_NOT_SUPPORTED)


def DateFromTicks(ticks: float) -> datetime.date:
    """
    Constructor mandated by the database API for creating a date value given
    the number of seconds since the epoch (January 1, 1970). This is equivalent
    to using datetime.date.fromtimestamp() and that should be used instead.
    """
    return datetime.date.fromtimestamp(ticks)


def TimeFromTicks(ticks: float) -> None:
    """
    Constructor mandated by the database API for creating a time value given
    the number of seconds since the epoch (January 1, 1970). Since Oracle
    doesn't support time only values, an exception is raised when this method
    is called.
    """
    errors._raise_err(errors.ERR_TIME_NOT_SUPPORTED)


def TimestampFromTicks(ticks: float) -> datetime.datetime:
    """
    Constructor mandated by the database API for creating a timestamp value
    given the number of seconds since the epoch (January 1, 1970). This is
    equivalent to using datetime.datetime.fromtimestamp() and that should be
    used instead.
    """
    return datetime.datetime.fromtimestamp(ticks)



#------------------------------------------------------------------------------
# Copyright (c) 2021, 2023, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
3600 - Module for testing the conversions of outputtype handler.
"""

import datetime
import decimal
import unittest

import oracledb
import test_env

class TestCase(test_env.BaseTestCase):

    def __test_type_handler(self, input_type, output_type, in_value,
                            expected_out_value):
        def type_handler(cursor, name, default_type, size, precision, scale):
            return cursor.var(output_type, arraysize=cursor.arraysize)
        self.cursor.outputtypehandler = type_handler
        var = self.cursor.var(input_type)
        var.setvalue(0, in_value)
        self.cursor.execute("select :1 from dual", [var])
        fetched_value, = self.cursor.fetchone()
        self.assertEqual(type(fetched_value), type(expected_out_value))
        self.assertEqual(fetched_value, expected_out_value)

    def __test_type_handler_lob(self, lob_type, output_type):
        db_type = getattr(oracledb, lob_type)
        def type_handler(cursor, name, default_type, size, precision, scale):
            if default_type == db_type:
                return cursor.var(output_type, arraysize=cursor.arraysize)
        self.cursor.outputtypehandler = type_handler
        in_value = f"Some {lob_type} data"
        if lob_type == "BLOB":
            in_value = in_value.encode()
        self.cursor.execute(f"truncate table Test{lob_type}s")
        self.cursor.execute(f"""
                insert into Test{lob_type}s
                (IntCol, {lob_type}Col)
                values(1, :val)""",
                val=in_value)
        self.connection.commit()
        self.cursor.execute(f"""
                select
                    {lob_type}Col,
                    IntCol,
                    {lob_type}Col
                from Test{lob_type}s""")
        self.assertEqual(self.cursor.fetchone(), (in_value, 1, in_value))

    def setUp(self):
        super().setUp()
        stmt = "ALTER SESSION SET NLS_DATE_FORMAT = 'YYYY-MM-DD HH24:MI:SS'" \
               "NLS_TIMESTAMP_FORMAT = 'YYYY-MM-DD HH24:MI:SS.FF6'" \
               "NLS_TIMESTAMP_TZ_FORMAT = 'YYYY-MM-DD HH24:MI:SS.FF6'" \
               "time_zone='Europe/London'"
        self.cursor.execute(stmt)

    def test_3600_VARCHAR_to_NUMBER(self):
        "3600 - output type handler conversion from VARCHAR to NUMBER"
        self.__test_type_handler(oracledb.DB_TYPE_VARCHAR,
                                 oracledb.DB_TYPE_NUMBER, "31.5", 31.5)

    def test_3601_CHAR_to_NUMBER(self):
        "3601 - output type handler conversion from CHAR to NUMBER"
        self.__test_type_handler(oracledb.DB_TYPE_CHAR,
                                 oracledb.DB_TYPE_NUMBER, "31.5", 31.5)

    def test_3602_LONG_to_NUMBER(self):
        "3602 - output type handler conversion from LONG to NUMBER"
        self.__test_type_handler(oracledb.DB_TYPE_LONG,
                                 oracledb.DB_TYPE_NUMBER, "31.5", 31.5)

    def test_3603_BINT_to_NUMBER(self):
        "3603 - test output type handler conversion from INTEGER to NUMBER"
        self.__test_type_handler(oracledb.DB_TYPE_BINARY_INTEGER,
                                 oracledb.DB_TYPE_NUMBER, 31, 31.0)

    def test_3604_VARCHAR_to_BINT(self):
        "3604 - output type handler conversion from VARCHAR to INTEGER"
        self.__test_type_handler(oracledb.DB_TYPE_VARCHAR,
                                 oracledb.DB_TYPE_BINARY_INTEGER, "31.5", 31)

    def test_3605_CHAR_to_BINT(self):
        "3605 - output type handler conversion from CHAR to INTEGER"
        self.__test_type_handler(oracledb.DB_TYPE_CHAR,
                                 oracledb.DB_TYPE_BINARY_INTEGER, "31.5", 31)

    def test_3606_LONG_to_BINT(self):
        "3606 - output type handler conversion from LONG to INTEGER"
        self.__test_type_handler(oracledb.DB_TYPE_LONG,
                                 oracledb.DB_TYPE_BINARY_INTEGER, "31.5", 31)

    def test_3607_NUMBER_to_BINT(self):
        "3607 - output type handler conversion from NUMBER to INTEGER"
        self.__test_type_handler(oracledb.DB_TYPE_NUMBER,
                                 oracledb.DB_TYPE_BINARY_INTEGER, 31.5, 31)

    def test_3608_BINARY_DOUBLE_to_BINT(self):
        "3608 - output type handler conversion from DOUBLE to INTEGER"
        self.__test_type_handler(oracledb.DB_TYPE_BINARY_DOUBLE,
                                 oracledb.DB_TYPE_BINARY_INTEGER, 31.5, 31)

    def test_3609_BINARY_FLOAT_to_BINT(self):
        "3609 - output type handler conversion from FLOAT to INTEGER"
        self.__test_type_handler(oracledb.DB_TYPE_BINARY_FLOAT,
                                 oracledb.DB_TYPE_BINARY_INTEGER, 31.5, 31)

    def test_3610_DATE_to_VARCHAR(self):
        "3610 - output type handler conversion from DATE to VARCHAR"
        in_val = datetime.date(2021, 2, 1)
        out_val = "2021-02-01 00:00:00"
        self.__test_type_handler(oracledb.DB_TYPE_DATE,
                                 oracledb.DB_TYPE_VARCHAR, in_val, out_val)

    def test_3611_DATE_to_CHAR(self):
        "3611 - output type handler conversion from DATE to CHAR"
        in_val = datetime.date(2021, 2, 1)
        out_val = "2021-02-01 00:00:00"
        self.__test_type_handler(oracledb.DB_TYPE_DATE,
                                 oracledb.DB_TYPE_CHAR, in_val, out_val)

    def test_3612_DATE_to_LONG(self):
        "3612 - output type handler conversion from DATE to LONG"
        in_val = datetime.date(2021, 2, 1)
        out_val = "2021-02-01 00:00:00"
        self.__test_type_handler(oracledb.DB_TYPE_DATE,
                                 oracledb.DB_TYPE_LONG, in_val , out_val)

    def test_3613_NUMBER_to_VARCHAR(self):
        "3613 - output type handler conversion from NUMBER to VARCHAR"
        self.__test_type_handler(oracledb.DB_TYPE_NUMBER,
                                 oracledb.DB_TYPE_VARCHAR, 31.5, "31.5")
        self.__test_type_handler(oracledb.DB_TYPE_NUMBER,
                                 oracledb.DB_TYPE_VARCHAR, 0, "0")

    def test_3614_NUMBER_to_CHAR(self):
        "3614 - output type handler conversion from NUMBER to CHAR"
        self.__test_type_handler(oracledb.DB_TYPE_NUMBER,
                                 oracledb.DB_TYPE_CHAR, 31.5, "31.5")
        self.__test_type_handler(oracledb.DB_TYPE_NUMBER,
                                 oracledb.DB_TYPE_CHAR, 0, "0")

    def test_3615_NUMBER_to_LONG(self):
        "3615 - output type handler conversion from NUMBER to LONG"
        self.__test_type_handler(oracledb.DB_TYPE_NUMBER,
                                 oracledb.DB_TYPE_LONG, 31.5, "31.5")

    def test_3616_INTERVAL_to_VARCHAR(self):
        "3616 - output type handler conversion from INTERVAL to VARCHAR"
        in_val = datetime.timedelta(days=-1, seconds=86314,
                                    microseconds=431152)
        if test_env.get_is_thin():
            out_val = str(in_val)
        else:
            out_val = "-000000001 23:58:34.431152000"
        self.__test_type_handler(oracledb.DB_TYPE_INTERVAL_DS,
                                 oracledb.DB_TYPE_VARCHAR, in_val, out_val)

    def test_3617_INTERVAL_to_CHAR(self):
        "3617 - output type handler conversion from INTERVAL to CHAR"
        in_val = datetime.timedelta(days=-1, seconds=86314,
                                    microseconds=431152)
        if test_env.get_is_thin():
            out_val = str(in_val)
        else:
            out_val = "-000000001 23:58:34.431152000"
        self.__test_type_handler(oracledb.DB_TYPE_INTERVAL_DS,
                                 oracledb.DB_TYPE_CHAR, in_val, out_val)

    def test_3618_INTERVAL_to_LONG(self):
        "3618 - output type handler conversion from INTERVAL to LONG"
        in_val = datetime.timedelta(days=-1, seconds=86314,
                                    microseconds=431152)
        if test_env.get_is_thin():
            out_val = str(in_val)
        else:
            out_val = "-000000001 23:58:34.431152000"
        self.__test_type_handler(oracledb.DB_TYPE_INTERVAL_DS,
                                 oracledb.DB_TYPE_LONG, in_val, out_val)

    def test_3619_TIMESTAMP_to_VARCHAR(self):
        "3619 - output type handler conversion from TIMESTAMP to VARCHAR"
        in_val = datetime.datetime(2002, 12, 17, 1, 2, 16, 400000)
        self.__test_type_handler(oracledb.DB_TYPE_TIMESTAMP,
                                 oracledb.DB_TYPE_VARCHAR, in_val, str(in_val))

    def test_3620_TIMESTAMP_to_CHAR(self):
        "3620 - output type handler conversion from TIMESTAMP to CHAR"
        in_val = datetime.datetime(2002, 12, 17, 1, 2, 16, 400000)
        self.__test_type_handler(oracledb.DB_TYPE_TIMESTAMP,
                                 oracledb.DB_TYPE_CHAR, in_val, str(in_val))

    def test_3621_TIMESTAMP_to_LONG(self):
        "3621 - output type handler conversion from TIMESTAMP to LONG"
        in_val = datetime.datetime(2002, 12, 17, 1, 2, 16, 400000)
        self.__test_type_handler(oracledb.DB_TYPE_TIMESTAMP,
                                 oracledb.DB_TYPE_LONG, in_val, str(in_val))

    def test_3622_TIMESTAMP_TZ_to_VARCHAR(self):
        "3622 - output type handler conversion from TIMESTAMP_TZ to VARCHAR"
        in_val = datetime.datetime(2002, 12, 17, 1, 2, 16, 400000)
        self.__test_type_handler(oracledb.DB_TYPE_TIMESTAMP_TZ,
                                 oracledb.DB_TYPE_VARCHAR, in_val, str(in_val))

    def test_3623_TIMESTAMP_TZ_to_CHAR(self):
        "3623 - output type handler conversion from TIMESTAMP_TZ to CHAR"
        in_val = datetime.datetime(2002, 12, 17, 1, 2, 16, 400000)
        self.__test_type_handler(oracledb.DB_TYPE_TIMESTAMP_TZ,
                                 oracledb.DB_TYPE_CHAR, in_val, str(in_val))

    def test_3624_TIMESTAMP_TZ_to_LONG(self):
        "3624 - output type handler conversion from TIMESTAMP_TZ to LONG"
        in_val = datetime.datetime(2002, 12, 17, 1, 2, 16, 400000)
        self.__test_type_handler(oracledb.DB_TYPE_TIMESTAMP_TZ,
                                 oracledb.DB_TYPE_LONG, in_val, str(in_val))

    def test_3625_TIMESTAMP_LTZ_to_VARCHAR(self):
        "3625 - output type handler conversion from TIMESTAMP_LTZ to VARCHAR"
        in_val = datetime.datetime(2002, 12, 17, 1, 2, 16, 400000)
        self.__test_type_handler(oracledb.DB_TYPE_TIMESTAMP_LTZ,
                                 oracledb.DB_TYPE_VARCHAR, in_val, str(in_val))

    def test_3626_TIMESTAMP_LTZ_to_CHAR(self):
        "3626 - output type handler conversion from TIMESTAMP_LTZ to CHAR"
        in_val = datetime.datetime(2002, 12, 17, 1, 2, 16, 400000)
        self.__test_type_handler(oracledb.DB_TYPE_TIMESTAMP_LTZ,
                                 oracledb.DB_TYPE_CHAR, in_val, str(in_val))

    def test_3627_TIMESTAMP_LTZ_to_LONG(self):
        "3627 - output type handler conversion from TIMESTAMP_LTZ to LONG"
        in_val = datetime.datetime(2002, 12, 17, 1, 2, 16, 400000)
        self.__test_type_handler(oracledb.DB_TYPE_TIMESTAMP_LTZ,
                                 oracledb.DB_TYPE_LONG, in_val, str(in_val))

    def test_3628_BINT_to_VARCHAR(self):
        "3628 - output type handler conversion from INTEGER to VARCHAR"
        self.__test_type_handler(oracledb.DB_TYPE_BINARY_INTEGER,
                                 oracledb.DB_TYPE_VARCHAR, 31, "31")

    def test_3629_BINT_to_CHAR(self):
        "3629 - output type handler conversion from INTEGER to CHAR"
        self.__test_type_handler(oracledb.DB_TYPE_BINARY_INTEGER,
                                 oracledb.DB_TYPE_CHAR, 31, "31")

    def test_3630_BINT_to_LONG(self):
        "3630 - output type handler conversion from INTEGER to LONG"
        self.__test_type_handler(oracledb.DB_TYPE_BINARY_INTEGER,
                                 oracledb.DB_TYPE_LONG, 31, "31")

    def test_3631_NUMBER_to_BINARY_DOUBLE(self):
        "3631 - output type handler conversion from NUMBER to DOUBLE"
        self.__test_type_handler(oracledb.DB_TYPE_NUMBER,
                                 oracledb.DB_TYPE_BINARY_DOUBLE, 31.5, 31.5)

    def test_3632_BINARY_FLOAT_to_BINARY_DOUBLE(self):
        "3632 - output type handler conversion from FLOAT to DOUBLE"
        self.__test_type_handler(oracledb.DB_TYPE_BINARY_FLOAT,
                                 oracledb.DB_TYPE_BINARY_DOUBLE, 31.5, 31.5)

    def test_3633_VARCHAR_to_BINARY_DOUBLE(self):
        "3633 - output type handler conversion from VARCHAR to DOUBLE"
        self.__test_type_handler(oracledb.DB_TYPE_VARCHAR,
                                 oracledb.DB_TYPE_BINARY_DOUBLE, "31.5", 31.5)

    def test_3634_CHAR_to_BINARY_DOUBLE(self):
        "3634 - output type handler conversion from CHAR to DOUBLE"
        self.__test_type_handler(oracledb.DB_TYPE_CHAR,
                                 oracledb.DB_TYPE_BINARY_DOUBLE, "31.5", 31.5)

    def test_3635_LONG_to_BINARY_DOUBLE(self):
        "3635 - output type handler conversion from LONG to DOUBLE"
        self.__test_type_handler(oracledb.DB_TYPE_LONG,
                                 oracledb.DB_TYPE_BINARY_DOUBLE, "31.5", 31.5)

    def test_3636_NUMBER_to_BINARY_FLOAT(self):
        "3636 - output type handler conversion from NUMBER to FLOAT"
        self.__test_type_handler(oracledb.DB_TYPE_NUMBER,
                                 oracledb.DB_TYPE_BINARY_FLOAT, 31.5, 31.5)

    def test_3637_BINARY_DOUBLE_to_BINARY_FLOAT(self):
        "3637 - output type handler conversion from DOUBLE to FLOAT"
        self.__test_type_handler(oracledb.DB_TYPE_BINARY_DOUBLE,
                                 oracledb.DB_TYPE_BINARY_FLOAT, 31.5, 31.5)

    def test_3638_VARCHAR_to_BINARY_FLOAT(self):
        "3638 - output type handler conversion from VARCHAR to FLOAT"
        self.__test_type_handler(oracledb.DB_TYPE_VARCHAR,
                                 oracledb.DB_TYPE_BINARY_FLOAT, "31.5", 31.5)

    def test_3639_CHAR_to_BINARY_FLOAT(self):
        "3639 - output type handler conversion from CHAR to FLOAT"
        self.__test_type_handler(oracledb.DB_TYPE_CHAR,
                                 oracledb.DB_TYPE_BINARY_FLOAT, "31.5", 31.5)

    def test_3640_LONG_to_BINARY_FLOAT(self):
        "3640 - output type handler conversion from LONG to FLOAT"
        self.__test_type_handler(oracledb.DB_TYPE_LONG,
                                 oracledb.DB_TYPE_BINARY_FLOAT, "31.5", 31.5)

    def test_3641_VARCHAR_to_CHAR(self):
        "3641 - output type handler conversion from VARCHAR to CHAR"
        self.__test_type_handler(oracledb.DB_TYPE_VARCHAR,
                                 oracledb.DB_TYPE_CHAR, "31.5", "31.5")

    def test_3642_VARCHAR_to_LONG(self):
        "3642 - output type handler conversion from VARCHAR to LONG"
        self.__test_type_handler(oracledb.DB_TYPE_VARCHAR,
                                 oracledb.DB_TYPE_LONG, "31.5", "31.5")

    def test_3643_LONG_to_VARCHAR(self):
        "3643 - output type handler conversion from LONG to VARCHAR"
        self.__test_type_handler(oracledb.DB_TYPE_LONG,
                                 oracledb.DB_TYPE_VARCHAR, "31.5", "31.5")

    def test_3644_LONG_to_CHAR(self):
        "3644 - output type handler conversion from LONG to CHAR"
        self.__test_type_handler(oracledb.DB_TYPE_LONG,
                                 oracledb.DB_TYPE_CHAR, "31.5", "31.5")

    def test_3645_CHAR_to_VARCHAR(self):
        "3645 - output type handler conversion from CHAR to VARCHAR"
        self.__test_type_handler(oracledb.DB_TYPE_CHAR,
                                 oracledb.DB_TYPE_VARCHAR, "31.5", "31.5")

    def test_3646_CHAR_to_LONG(self):
        "3646 - output type handler conversion from CHAR to LONG"
        self.__test_type_handler(oracledb.DB_TYPE_CHAR,
                                 oracledb.DB_TYPE_LONG, "31.5", "31.5")

    def test_3647_TIMESTAMP_to_TIMESTAMP_TZ(self):
        "3647 - output type handler conversion from TIMESTAMP to TIMESTAMP_TZ"
        val = datetime.datetime(2002, 12, 17, 0, 0, 16, 400000)
        self.__test_type_handler(oracledb.DB_TYPE_TIMESTAMP,
                                 oracledb.DB_TYPE_TIMESTAMP_TZ, val, val)

    def test_3648_TIMESTAMP_to_TIMESTAMP_LTZ(self):
        "3648 - output type handler conversion from TIMESTAMP to TIMESTAMP_LTZ"
        val = datetime.datetime(2002, 12, 17, 0, 0, 16, 400000)
        self.__test_type_handler(oracledb.DB_TYPE_TIMESTAMP,
                                 oracledb.DB_TYPE_TIMESTAMP_LTZ, val, val)

    def test_3649_TIMESTAMP_TZ_to_TIMESTAMP(self):
        "3649 - output type handler from conversion TIMESTAMP_TZ to TIMESTAMP"
        val = datetime.datetime(2002, 12, 17, 0, 0, 16, 400000)
        self.__test_type_handler(oracledb.DB_TYPE_TIMESTAMP_TZ,
                                 oracledb.DB_TYPE_TIMESTAMP, val, val)

    def test_3650_NUMBER_TO_DATE(self):
        "3650 - output type handler conversion from NUMBER to DATE is invalid"
        self.assertRaisesRegex(oracledb.DatabaseError,
                               "^DPY-4007:|^ORA-00932:",
                               self.__test_type_handler,
                               oracledb.DB_TYPE_NUMBER, oracledb.DB_TYPE_DATE,
                               3, 3)

    def test_3651_CLOB_TO_CHAR(self):
        "3651 - output type handler conversion from CLOB to CHAR"
        val = "Some Clob String"
        self.__test_type_handler(oracledb.DB_TYPE_CLOB,
                                 oracledb.DB_TYPE_CHAR, val, val)

    def test_3652_CLOB_TO_VARCHAR(self):
        "3652 - output type handler conversion from CLOB to VARCHAR"
        val = "Some Clob String"
        self.__test_type_handler(oracledb.DB_TYPE_CLOB,
                                 oracledb.DB_TYPE_VARCHAR, val, val)

    def test_3653_CLOB_TO_LONG(self):
        "3653 - output type handler conversion from CLOB to LONG"
        val = "Some Clob String"
        self.__test_type_handler(oracledb.DB_TYPE_CLOB,
                                 oracledb.DB_TYPE_LONG, val, val)

    def test_3654_BLOB_TO_RAW(self):
        "3654 - output type handler conversion from BLOB to RAW"
        val = b"Some binary data"
        self.__test_type_handler(oracledb.DB_TYPE_BLOB,
                                 oracledb.DB_TYPE_RAW, val, val)

    def test_3655_BLOB_TO_LONG_RAW(self):
        "3655 - output type handler conversion from BLOB to LONGRAW"
        val = b"Some binary data"
        self.__test_type_handler(oracledb.DB_TYPE_BLOB,
                                 oracledb.DB_TYPE_LONG_RAW, val, val)

    def test_3656_BLOB_TO_LONG_RAW(self):
        "3656 - output type handler conversion from permanent BLOBs to LONG_RAW"
        self.__test_type_handler_lob("BLOB", oracledb.DB_TYPE_LONG_RAW)

    def test_3657_BLOB_TO_RAW(self):
        "3657 - output type handler conversion from permanent BLOBs to RAW"
        self.__test_type_handler_lob("BLOB", oracledb.DB_TYPE_RAW)

    def test_3658_CLOB_TO_VARCHAR(self):
        "3658 - output type handler conversion from permanent CLOBs to VARCHAR"
        self.__test_type_handler_lob("CLOB", oracledb.DB_TYPE_VARCHAR)

    def test_3659_CLOB_TO_CHAR(self):
        "3659 - output type handler conversion from permanent CLOBs to CHAR"
        self.__test_type_handler_lob("CLOB", oracledb.DB_TYPE_CHAR)

    def test_3660_CLOB_TO_LONG(self):
        "3660 - output type handler conversion from permanent CLOBs to LONG"
        self.__test_type_handler_lob("CLOB", oracledb.DB_TYPE_LONG)

    def test_3661_NCLOB_TO_CHAR(self):
        "3661 - output type handler conversion from NCLOB to CHAR"
        val = "Some nclob data"
        self.__test_type_handler(oracledb.DB_TYPE_NCLOB,
                                 oracledb.DB_TYPE_CHAR, val, val)

    def test_3662_NCLOB_TO_VARCHAR(self):
        "3662 - output type handler conversion from NCLOB to VARCHAR"
        val = "Some nclob data"
        self.__test_type_handler(oracledb.DB_TYPE_NCLOB,
                                 oracledb.DB_TYPE_VARCHAR, val, val)

    def test_3663_NCLOB_TO_LONG(self):
        "3663 - output type handler conversion from NCLOB to LONG"
        val = "Some nclob data"
        self.__test_type_handler(oracledb.DB_TYPE_NCLOB,
                                 oracledb.DB_TYPE_LONG, val, val)

    def test_3664_NCLOB_TO_VARCHAR(self):
        "3664 - output type handler conversion from permanent NCLOBs to VARCHAR"
        self.__test_type_handler_lob("NCLOB", oracledb.DB_TYPE_VARCHAR)

    def test_3665_NCLOB_TO_CHAR(self):
        "3665 - output type handler conversion from permanent NCLOBs to CHAR"
        self.__test_type_handler_lob("NCLOB", oracledb.DB_TYPE_CHAR)

    def test_3666_NCLOB_TO_LONG(self):
        "3666 - output type handler conversion from permanent NCLOBs to LONG"
        self.__test_type_handler_lob("NCLOB", oracledb.DB_TYPE_LONG)

    def test_3667_NVARCHAR_to_VARCHAR(self):
        "3667 - output type handler conversion from NVARCHAR to VARCHAR"
        self.__test_type_handler(oracledb.DB_TYPE_NVARCHAR,
                                 oracledb.DB_TYPE_VARCHAR, "31.5", "31.5")

    def test_3668_VARCHAR_to_NVARCHAR(self):
        "3668 - output type handler conversion from VARCHAR to NVARCHAR"
        self.__test_type_handler(oracledb.DB_TYPE_VARCHAR,
                                 oracledb.DB_TYPE_NVARCHAR, "31.5", "31.5")

    def test_3669_NCHAR_to_CHAR(self):
        "3669 - output type handler conversion from NCHAR to CHAR"
        self.__test_type_handler(oracledb.DB_TYPE_NCHAR,
                                 oracledb.DB_TYPE_CHAR, "31.5", "31.5")

    def test_3670_CHAR_to_NCHAR(self):
        "3670 - output type handler conversion from CHAR to NCHAR"
        self.__test_type_handler(oracledb.DB_TYPE_CHAR,
                                 oracledb.DB_TYPE_NCHAR, "31.5", "31.5")

    def test_3671_incorrect_arraysize(self):
        "3671 - execute raises an error if an incorrect arraysize is used"
        def type_handler(cursor, name, default_type, size, precision, scale):
            return cursor.var(str)
        cursor = self.connection.cursor()
        cursor.arraysize = 100
        cursor.outputtypehandler = type_handler
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2016:",
                               cursor.execute, "select :1 from dual", [5])

    def test_3672_incorrect_outputtypehandler_return_type(self):
        "3672 - execute raises an error if a var is not returned"
        def type_handler(cursor, name, default_type, size,
                         precision, scale):
            return "incorrect_return"
        cursor = self.connection.cursor()
        cursor.outputtypehandler = type_handler
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2015:",
                               cursor.execute, "select :1 from dual", [5])

    def test_3673_NUMBER_to_DECIMAL(self):
        "3673 - output type handler conversion from NUMBER to decimal.Decimal"
        self.__test_type_handler(oracledb.DB_TYPE_NUMBER, decimal.Decimal,
                                 31.5, decimal.Decimal("31.5"))
        self.__test_type_handler(oracledb.DB_TYPE_NUMBER, decimal.Decimal, 0,
                                 decimal.Decimal("0"))

    def test_3674_cursor_description_unchanged(self):
        "3674 - use of output type handler does not affect description"
        def type_handler(cursor, name, default_type, size,
                         precision, scale):
            return cursor.var(str, arraysize=cursor.arraysize)
        with self.connection.cursor() as cursor:
            cursor.execute("select user from dual")
            desc_before = cursor.description
        with self.connection.cursor() as cursor:
            cursor.outputtypehandler = type_handler
            cursor.execute("select user from dual")
            self.assertEqual(cursor.description, desc_before)

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# dbobject.pyx
#
# Cython file defining the base DbObjectType, DbObjectAttr and DbObject
# implementation classes (embedded in base_impl.pyx).
#------------------------------------------------------------------------------

cdef class BaseDbObjectImpl:

    def append(self, object value):
        """
        Appends a value to the collection after first checking to see if the
        value is acceptable.
        """
        cdef BaseConnImpl conn_impl = self.type._conn_impl
        value = conn_impl._check_value(self.type.element_dbtype,
                                       self.type.element_objtype, value, NULL)
        self.append_checked(value)

    @utils.CheckImpls("appending a value to a collection")
    def append_checked(self, object value):
        pass

    @utils.CheckImpls("creating a copy of an object")
    def copy(self):
        pass

    @utils.CheckImpls("deleting an element in a collection")
    def delete_by_index(self, int32_t index):
        pass

    @utils.CheckImpls("determining if an entry exists in a collection")
    def exists_by_index(self, int32_t index):
        pass

    @utils.CheckImpls("getting an attribute value")
    def get_attr_value(self, BaseDbObjectAttrImpl attr):
        pass

    @utils.CheckImpls("getting an element of a collection")
    def get_element_by_index(self, int32_t index):
        pass

    @utils.CheckImpls("getting the first index of a collection")
    def get_first_index(self):
        pass

    @utils.CheckImpls("getting the last index of a collection")
    def get_last_index(self):
        pass

    @utils.CheckImpls("getting the next index of a collection")
    def get_next_index(self, int32_t index):
        pass

    @utils.CheckImpls("getting the previous index of a collection")
    def get_prev_index(self, int32_t index):
        pass

    @utils.CheckImpls("getting the size of a collection")
    def get_size(self):
        pass

    def set_attr_value(self, BaseDbObjectAttrImpl attr, object value):
        """
        Sets the attribute value after first checking to see if the value is
        acceptable.
        """
        cdef BaseConnImpl conn_impl = self.type._conn_impl
        value = conn_impl._check_value(attr.dbtype, attr.objtype, value, NULL)
        self.set_attr_value_checked(attr, value)

    @utils.CheckImpls("setting an attribute value")
    def set_attr_value_checked(self, BaseDbObjectAttrImpl attr, object value):
        pass

    def set_element_by_index(self, int32_t index, object value):
        """
        Sets the element value after first checking to see if the value is
        acceptable.
        """
        cdef BaseConnImpl conn_impl = self.type._conn_impl
        value = conn_impl._check_value(self.type.element_dbtype,
                                       self.type.element_objtype, value, NULL)
        self.set_element_by_index_checked(index, value)

    @utils.CheckImpls("setting an element of a collection")
    def set_element_by_index_checked(self, int32_t index, object value):
        pass

    @utils.CheckImpls("trimming elements from a collection")
    def trim(self, int32_t num_to_trim):
        pass


cdef class BaseDbObjectAttrImpl:
    pass


cdef class BaseDbObjectTypeImpl:

    def __eq__(self, other):
        if isinstance(other, BaseDbObjectTypeImpl):
            return other._conn_impl is self._conn_impl \
                    and other.schema == self.schema \
                    and other.name == self.name
        return NotImplemented

    @utils.CheckImpls("creating a new object")
    def create_new_object(self):
        pass



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
1000 - Module for testing top-level module methods
"""

import datetime
import time

import oracledb
import test_env

class TestCase(test_env.BaseTestCase):
    requires_connection = False

    def test_1000_date_from_ticks(self):
        "1000 - test DateFromTicks()"
        today = datetime.datetime.today()
        timestamp = time.mktime(today.timetuple())
        date = oracledb.DateFromTicks(timestamp)
        self.assertEqual(date, today.date())

    def test_1001_future_obj(self):
        "1001 - test management of __future__ object"
        self.assertEqual(oracledb.__future__.dummy, None)
        oracledb.__future__.dummy = "Unimportant"
        self.assertEqual(oracledb.__future__.dummy, None)

    def test_1002_timestamp_from_ticks(self):
        "1002 - test TimestampFromTicks()"
        timestamp = time.mktime(datetime.datetime.today().timetuple())
        today = datetime.datetime.fromtimestamp(timestamp)
        date = oracledb.TimestampFromTicks(timestamp)
        self.assertEqual(date, today)

    def test_1003_unsupported_functions(self):
        "1003 - test unsupported time functions"
        self.assertRaisesRegex(oracledb.NotSupportedError, "^DPY-3000:",
                               oracledb.Time, 12, 0, 0)
        self.assertRaisesRegex(oracledb.NotSupportedError, "^DPY-3000:",
                               oracledb.TimeFromTicks, 100)

    def test_1004_makedsn(self):
        "1004 - test makedsn() with valid arguments"
        format_string = "(DESCRIPTION=(ADDRESS=(PROTOCOL=TCP)" + \
                        "(HOST=%s)(PORT=%d))(CONNECT_DATA=(SID=%s)))"
        args = ("hostname", 1521, "TEST")
        result = oracledb.makedsn(*args)
        self.assertEqual(result, format_string % args)

    def test_1005_makedsn_invalid_args(self):
        "1005 - test makedsn() with invalid arguments"
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2020:",
                               oracledb.makedsn, host="(invalid)", port=1521)
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2020:",
                               oracledb.makedsn, host="host", port=1521,
                               sid="(invalid)")
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2020:",
                               oracledb.makedsn, host="host", port=1521,
                               service_name="(invalid)")
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2020:",
                               oracledb.makedsn, host="host", port=1521,
                               region="(invalid)")
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2020:",
                               oracledb.makedsn, host="host", port=1521,
                               sharding_key="(invalid)")
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2020:",
                               oracledb.makedsn, host="host", port=1521,
                               super_sharding_key="(invalid)")

    def test_1006_aliases(self):
        "1006 - test aliases match"

        # database type aliases
        self.assertIs(oracledb.BFILE, oracledb.DB_TYPE_BFILE)
        self.assertIs(oracledb.BLOB, oracledb.DB_TYPE_BLOB)
        self.assertIs(oracledb.BOOLEAN, oracledb.DB_TYPE_BOOLEAN)
        self.assertIs(oracledb.CLOB, oracledb.DB_TYPE_CLOB)
        self.assertIs(oracledb.CURSOR, oracledb.DB_TYPE_CURSOR)
        self.assertIs(oracledb.FIXED_CHAR, oracledb.DB_TYPE_CHAR)
        self.assertIs(oracledb.FIXED_NCHAR, oracledb.DB_TYPE_NCHAR)
        self.assertIs(oracledb.INTERVAL, oracledb.DB_TYPE_INTERVAL_DS)
        self.assertIs(oracledb.LONG_BINARY, oracledb.DB_TYPE_LONG_RAW)
        self.assertIs(oracledb.LONG_STRING, oracledb.DB_TYPE_LONG)
        self.assertIs(oracledb.NATIVE_INT, oracledb.DB_TYPE_BINARY_INTEGER)
        self.assertIs(oracledb.NATIVE_FLOAT, oracledb.DB_TYPE_BINARY_DOUBLE)
        self.assertIs(oracledb.NCHAR, oracledb.DB_TYPE_NVARCHAR)
        self.assertIs(oracledb.NCLOB, oracledb.DB_TYPE_NCLOB)
        self.assertIs(oracledb.OBJECT, oracledb.DB_TYPE_OBJECT)
        self.assertIs(oracledb.TIMESTAMP, oracledb.DB_TYPE_TIMESTAMP)

        # type aliases
        self.assertIs(oracledb.ObjectType, oracledb.DbObjectType)
        self.assertIs(oracledb.Object, oracledb.DbObject)
        self.assertIs(oracledb.SessionPool, oracledb.ConnectionPool)

        # authentication mode aliases
        self.assertIs(oracledb.DEFAULT_AUTH, oracledb.AUTH_MODE_DEFAULT)
        self.assertIs(oracledb.SYSASM, oracledb.AUTH_MODE_SYSASM)
        self.assertIs(oracledb.SYSBKP, oracledb.AUTH_MODE_SYSBKP)
        self.assertIs(oracledb.SYSDBA, oracledb.AUTH_MODE_SYSDBA)
        self.assertIs(oracledb.SYSDGD, oracledb.AUTH_MODE_SYSDGD)
        self.assertIs(oracledb.SYSKMT, oracledb.AUTH_MODE_SYSKMT)
        self.assertIs(oracledb.SYSOPER, oracledb.AUTH_MODE_SYSOPER)
        self.assertIs(oracledb.SYSRAC, oracledb.AUTH_MODE_SYSRAC)
        self.assertIs(oracledb.PRELIM_AUTH, oracledb.AUTH_MODE_PRELIM)

        # pool "get" mode aliases
        self.assertIs(oracledb.SPOOL_ATTRVAL_WAIT, oracledb.POOL_GETMODE_WAIT)
        self.assertIs(oracledb.SPOOL_ATTRVAL_NOWAIT,
                      oracledb.POOL_GETMODE_NOWAIT)
        self.assertIs(oracledb.SPOOL_ATTRVAL_FORCEGET,
                      oracledb.POOL_GETMODE_FORCEGET)
        self.assertIs(oracledb.SPOOL_ATTRVAL_TIMEDWAIT,
                      oracledb.POOL_GETMODE_TIMEDWAIT)

        # purity aliases
        self.assertIs(oracledb.ATTR_PURITY_DEFAULT, oracledb.PURITY_DEFAULT)
        self.assertIs(oracledb.ATTR_PURITY_NEW, oracledb.PURITY_NEW)
        self.assertIs(oracledb.ATTR_PURITY_SELF, oracledb.PURITY_SELF)

        # other aliases
        self.assertIs(oracledb.SUBSCR_PROTO_OCI,
                      oracledb.SUBSCR_PROTO_CALLBACK)
        self.assertIs(oracledb.version, oracledb.__version__)

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# thin_impl.pyx
#
# Cython file for communicating with the server directly without the use of
# any Oracle Client library.
#------------------------------------------------------------------------------

# cython: language_level=3

cimport cython
cimport cpython
cimport cpython.datetime as cydatetime
cimport cpython.ref

from libc.stdint cimport int8_t, int16_t, int32_t, int64_t
from libc.stdint cimport uint8_t, uint16_t, uint32_t, uint64_t
from libc.stdint cimport UINT8_MAX, UINT16_MAX, UINT32_MAX, UINT64_MAX
from libc.string cimport memcpy, memset
from cpython cimport array

import array
import base64
import collections
import datetime
import decimal
import getpass
import hashlib
import os
import socket
import re
import secrets
import select
import ssl
import subprocess
import sys
import threading
import time

try:
    import certifi
except ImportError:
    certifi = None
macos_certs = None

cydatetime.import_datetime()

from . import __version__ as VERSION

from . import constants, errors, exceptions
from .defaults import defaults

from .base_impl cimport get_exception_class, NUM_TYPE_FLOAT
from .base_impl cimport NUM_TYPE_INT, NUM_TYPE_DECIMAL, NUM_TYPE_STR
from .base_impl cimport BaseConnImpl, BaseCursorImpl, BaseVarImpl, DbType
from .base_impl cimport BaseLobImpl, BasePoolImpl, FetchInfo
from .base_impl cimport Address, AddressList, Description, DescriptionList
from .base_impl cimport ConnectParamsImpl, PoolParamsImpl, BaseDbObjectAttrImpl
from .base_impl cimport BaseDbObjectImpl, BaseDbObjectTypeImpl
from .base_impl import DB_TYPE_BLOB, DB_TYPE_CLOB, DB_TYPE_NCLOB
from .base_impl import DB_TYPE_BINARY_INTEGER, DB_TYPE_CURSOR, DB_TYPE_OBJECT

ctypedef unsigned char char_type

cdef type PY_TYPE_DATE = datetime.date
cdef type PY_TYPE_DATETIME = datetime.datetime
cdef type PY_TYPE_DECIMAL = decimal.Decimal
cdef type PY_TYPE_DB_OBJECT
cdef type PY_TYPE_LOB
cdef type PY_TYPE_TIMEDELTA = datetime.timedelta

cdef bint HAS_CRYPTOGRAPHY = True

include "impl/thin/constants.pxi"
include "impl/thin/utils.pyx"
include "impl/thin/crypto.pyx"
include "impl/thin/capabilities.pyx"
include "impl/thin/buffer.pyx"
include "impl/thin/packet.pyx"
include "impl/thin/data_types.pyx"
include "impl/thin/messages.pyx"
include "impl/thin/protocol.pyx"
include "impl/thin/connection.pyx"
include "impl/thin/statement.pyx"
include "impl/thin/cursor.pyx"
include "impl/thin/var.pyx"
include "impl/thin/dbobject.pyx"
include "impl/thin/oson.pyx"
include "impl/thin/lob.pyx"
include "impl/thin/pool.pyx"
include "impl/thin/conversions.pyx"



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# pool.pyx
#
# Cython file defining the thick Pool implementation class (embedded in
# thick_impl.pyx).
#------------------------------------------------------------------------------

cdef int _token_callback_handler(void *context,
                                 dpiAccessToken *refresh_token) with gil:
    cdef:
        ThickPoolImpl pool_impl = <object> context
    pool_impl._token_handler(refresh_token, pool_impl.connect_params)


cdef class ThickPoolImpl(BasePoolImpl):
    cdef:
        dpiPool *_handle

    def __init__(self, str dsn, PoolParamsImpl params):
        cdef:
            bytes edition_bytes, user_bytes, password_bytes, dsn_bytes
            uint32_t password_len = 0, user_len = 0, dsn_len = 0
            bytes session_callback_bytes, name_bytes
            dpiCommonCreateParams common_params
            dpiPoolCreateParams create_params
            const char *password_ptr = NULL
            const char *user_ptr = NULL
            const char *dsn_ptr = NULL
            bytes token_bytes, private_key_bytes
            uint32_t token_len = 0, private_key_len = 0
            const char *token_ptr = NULL
            const char *private_key_ptr = NULL
            dpiAccessToken access_token
            str token, private_key
            int status

        # save parameters
        self.connect_params = params
        self.username = params.user
        self.dsn = dsn
        self.min = params.min
        self.max = params.max
        self.increment = params.increment
        self.homogeneous = params.homogeneous

        # set up token parameters if provided
        if params._token is not None \
                or params.access_token_callback is not None:
            token = params._get_token()
            token_bytes = token.encode()
            token_ptr = token_bytes
            token_len = <uint32_t> len(token_bytes)
            private_key = params._get_private_key()
            if private_key is not None:
                private_key_bytes = private_key.encode()
                private_key_ptr = private_key_bytes
                private_key_len = <uint32_t> len(private_key_bytes)

        # set up common creation parameters
        if dpiContext_initCommonCreateParams(driver_context,
                                             &common_params) < 0:
            _raise_from_odpi()
        common_params.createMode |= DPI_MODE_CREATE_THREADED
        if params.events:
            common_params.createMode |= DPI_MODE_CREATE_EVENTS
        if params.edition is not None:
            edition_bytes = params.edition.encode()
            common_params.edition = edition_bytes
            common_params.editionLength = <uint32_t> len(edition_bytes)
        if params._token is not None:
            access_token.token = token_ptr
            access_token.tokenLength = token_len
            access_token.privateKey = private_key_ptr
            access_token.privateKeyLength = private_key_len
            common_params.accessToken = &access_token

        # set up pool creation parameters
        if dpiContext_initPoolCreateParams(driver_context, &create_params) < 0:
            _raise_from_odpi()
        create_params.minSessions = self.min
        create_params.maxSessions = self.max
        create_params.sessionIncrement = self.increment
        create_params.homogeneous = self.homogeneous
        create_params.getMode = params.getmode
        if params.session_callback is not None \
                and not callable(params.session_callback):
            session_callback_bytes = params.session_callback.encode()
            create_params.plsqlFixupCallback = session_callback_bytes
            create_params.plsqlFixupCallbackLength = \
                    <uint32_t> len(session_callback_bytes)
        if params.access_token_callback is not None:
            create_params.accessTokenCallback = _token_callback_handler
            create_params.accessTokenCallbackContext = <void*> self
        create_params.timeout = params.timeout
        create_params.waitTimeout = params.wait_timeout
        create_params.maxSessionsPerShard = params.max_sessions_per_shard
        create_params.maxLifetimeSession = params.max_lifetime_session
        create_params.pingInterval = params.ping_interval
        common_params.stmtCacheSize = params.stmtcachesize
        common_params.sodaMetadataCache = params.soda_metadata_cache
        create_params.externalAuth = params.externalauth

        # prepare user, password and DSN for use
        if self.username is not None:
            user_bytes = params.get_full_user().encode()
            user_ptr = user_bytes
            user_len = <uint32_t> len(user_bytes)
        password_bytes = params._get_password()
        if password_bytes is not None:
            password_ptr = password_bytes
            password_len = <uint32_t> len(password_bytes)
        if self.dsn is not None:
            dsn_bytes = self.dsn.encode()
            dsn_ptr = dsn_bytes
            dsn_len = <uint32_t> len(dsn_bytes)

        # create pool
        with nogil:
            status = dpiPool_create(driver_context, user_ptr, user_len,
                                    password_ptr, password_len, dsn_ptr,
                                    dsn_len, &common_params, &create_params,
                                    &self._handle)
        if status < 0:
            _raise_from_odpi()

        name_bytes = create_params.outPoolName[:create_params.outPoolNameLength]
        self.name = name_bytes.decode()

    def __dealloc__(self):
        if self._handle != NULL:
            dpiPool_release(self._handle)

    cdef object _token_handler(self, dpiAccessToken *access_token,
                               ConnectParamsImpl params):
        cdef:
            str token, private_key
            bytes token_bytes, private_key_bytes
            uint32_t token_len = 0, private_key_len = 0
            const char *token_ptr = NULL
            const char *private_key_ptr = NULL
        token = params._get_token()
        token_bytes = token.encode()
        token_ptr = token_bytes
        token_len = <uint32_t> len(token_bytes)
        private_key = params._get_private_key()
        if private_key is not None:
            private_key_bytes = private_key.encode()
            private_key_ptr = private_key_bytes
            private_key_len = <uint32_t> len(private_key_bytes)
        access_token.token = token_ptr
        access_token.tokenLength = token_len
        access_token.privateKey = private_key_ptr
        access_token.privateKeyLength = private_key_len

    def close(self, bint force):
        """
        Internal method for closing the pool.
        """
        cdef:
            uint32_t close_mode
            int status
        close_mode = DPI_MODE_POOL_CLOSE_FORCE if force \
                     else DPI_MODE_POOL_CLOSE_DEFAULT
        with nogil:
            status = dpiPool_close(self._handle, close_mode);
        if status < 0:
            _raise_from_odpi()

    def drop(self, ThickConnImpl conn_impl):
        """
        Internal method for dropping a connection from the pool.
        """
        cdef int status
        with nogil:
            status = dpiConn_close(conn_impl._handle, DPI_MODE_CONN_CLOSE_DROP,
                                   NULL, 0)
        if status < 0:
            _raise_from_odpi()
        dpiConn_release(conn_impl._handle)
        conn_impl._handle = NULL

    def get_busy_count(self):
        """
        Internal method for getting the number of busy connections in the pool.
        """
        cdef uint32_t value
        if dpiPool_getBusyCount(self._handle, &value) < 0:
            _raise_from_odpi()
        return value

    def get_getmode(self):
        """
        Internal method for getting the method by which connections are
        acquired from the pool.
        """
        cdef uint8_t value
        if dpiPool_getGetMode(self._handle, &value) < 0:
            _raise_from_odpi()
        return value

    def get_max_lifetime_session(self):
        """
        Internal method for getting the maximum lifetime of each session.
        """
        cdef uint32_t value
        if dpiPool_getMaxLifetimeSession(self._handle, &value) < 0:
            _raise_from_odpi()
        return value

    def get_max_sessions_per_shard(self):
        """
        Internal method for getting the maximum sessions per shard in the pool.
        """
        cdef uint32_t value
        if dpiPool_getMaxSessionsPerShard(self._handle, &value) < 0:
            _raise_from_odpi()
        return value

    def get_open_count(self):
        """
        Internal method for getting the number of connections in the pool.
        """
        cdef uint32_t value
        if dpiPool_getOpenCount(self._handle, &value) < 0:
            _raise_from_odpi()
        return value

    def get_ping_interval(self):
        """
        Internal method for getting the value of the pool-ping-interval.
        """
        cdef int value
        if dpiPool_getPingInterval(self._handle, &value) < 0:
            _raise_from_odpi
        return value

    def get_soda_metadata_cache(self):
        """
        Internal method for getting the value of soda metadata cache.
        """
        cdef bint value
        if dpiPool_getSodaMetadataCache(self._handle, &value) < 0:
            _raise_from_odpi
        return value

    def get_stmt_cache_size(self):
        """
        Internal method for getting the size of the statement cache.
        """
        cdef uint32_t value
        if dpiPool_getStmtCacheSize(self._handle, &value) < 0:
            _raise_from_odpi()
        return value

    def get_timeout(self):
        """
        Internal method for getting the timeout for idle sessions.
        """
        cdef uint32_t value
        if dpiPool_getTimeout(self._handle, &value) < 0:
            _raise_from_odpi()
        return value

    def get_wait_timeout(self):
        """
        Internal method for getting the wait timeout for acquiring sessions.
        """
        cdef uint32_t value
        if dpiPool_getWaitTimeout(self._handle, &value) < 0:
            _raise_from_odpi()
        return value

    def reconfigure(self, uint32_t min, uint32_t max, uint32_t increment):
        """
        Internal method for reconfiguring the size of the pool.
        """
        if dpiPool_reconfigure(self._handle, min, max, increment) < 0:
            _raise_from_odpi()
        self.min = min
        self.max = max
        self.increment = increment

    def set_getmode(self, uint8_t value):
        """
        Internal method for setting the method by which connections are
        acquired from the pool.
        """
        if dpiPool_setGetMode(self._handle, value) < 0:
            _raise_from_odpi()

    def set_max_lifetime_session(self, uint32_t value):
        """
        Internal method for setting the maximum lifetime of each session.
        """
        if dpiPool_setMaxLifetimeSession(self._handle, value) < 0:
            _raise_from_odpi()

    def set_max_sessions_per_shard(self, uint32_t value):
        """
        Internal method for setting the maximum sessions per shard in the pool.
        """
        if dpiPool_setMaxSessionsPerShard(self._handle, value) < 0:
            _raise_from_odpi()

    def set_ping_interval(self, int value):
        """
        Internal method for setting the value of the pool-ping-interval.
        """
        if dpiPool_setPingInterval(self._handle, value) < 0:
            _raise_from_odpi

    def set_soda_metadata_cache(self, bint value):
        """
        Internal method for enabling or disabling the soda metadata cache.
        """
        if dpiPool_setSodaMetadataCache(self._handle, value) < 0:
            _raise_from_odpi

    def set_stmt_cache_size(self, uint32_t value):
        """
        Internal method for setting the size of the statement cache.
        """
        if dpiPool_setStmtCacheSize(self._handle, value) < 0:
            _raise_from_odpi()

    def set_timeout(self, uint32_t value):
        """
        Internal method for setting the timeout for idle sessions.
        """
        if dpiPool_setTimeout(self._handle, value) < 0:
            _raise_from_odpi()

    def set_wait_timeout(self, uint32_t value):
        """
        Internal method for setting the wait timeout for acquiring sessions.
        """
        if dpiPool_setWaitTimeout(self._handle, value) < 0:
            _raise_from_odpi()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# drop_schema.py
#
# Drops the database objects used by the python-oracledb test suite.
#
# This script is also executed by the Python script create_schema.py for
# dropping the existing users and editions, if applicable, before creating the
# test schemas.
#------------------------------------------------------------------------------

import oracledb
import test_env

def drop_schema(conn):
    print("Dropping test schemas...")
    test_env.run_sql_script(conn, "drop_schema",
                            main_user=test_env.get_main_user(),
                            proxy_user=test_env.get_proxy_user())

if __name__ == "__main__":
    conn = test_env.get_admin_connection()
    drop_schema(conn)
    print("Done.")



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
1800 - Module for testing interval variables
"""

import datetime

import oracledb
import test_env

class TestCase(test_env.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.raw_data = []
        self.data_by_key = {}
        for i in range(1, 11):
            delta = datetime.timedelta(days=i, hours=i, minutes=i * 2,
                                       seconds=i * 3)
            if i % 2 == 0:
                nullable_delta = None
            else:
                nullable_delta = datetime.timedelta(days=i + 5, hours=i + 2,
                                                    minutes=i * 2 + 5,
                                                    seconds=i * 3 + 5)
            data_tuple = (i, delta, nullable_delta)
            self.raw_data.append(data_tuple)
            self.data_by_key[i] = data_tuple

    def test_1800_bind_interval(self):
        "1800 - test binding in an interval"
        self.cursor.setinputsizes(value=oracledb.DB_TYPE_INTERVAL_DS)
        value = datetime.timedelta(days=5, hours=5, minutes=10, seconds=15)
        self.cursor.execute("""
                select * from TestIntervals
                where IntervalCol = :value""",
                value=value)
        self.assertEqual(self.cursor.fetchall(), [self.data_by_key[5]])

    def test_1801_bind_null(self):
        "1801 - test binding in a null"
        self.cursor.setinputsizes(value=oracledb.DB_TYPE_INTERVAL_DS)
        self.cursor.execute("""
                select * from TestIntervals
                where IntervalCol = :value""",
                value=None)
        self.assertEqual(self.cursor.fetchall(), [])

    def test_1802_bind_out_set_input_sizes(self):
        "1802 - test binding out with set input sizes defined"
        bind_vars = \
                self.cursor.setinputsizes(value=oracledb.DB_TYPE_INTERVAL_DS)
        self.cursor.execute("""
                begin
                    :value := to_dsinterval('8 09:24:18.123789');
                end;""")
        expected_value = datetime.timedelta(days=8, hours=9, minutes=24,
                                            seconds=18, microseconds=123789)
        self.assertEqual(bind_vars["value"].getvalue(), expected_value)

    def test_1803_bind_in_out_set_input_sizes(self):
        "1803 - test binding in/out with set input sizes defined"
        bind_vars = \
                self.cursor.setinputsizes(value=oracledb.DB_TYPE_INTERVAL_DS)
        self.cursor.execute("""
                begin
                    :value := :value + to_dsinterval('5 08:30:00');
                end;""",
                value=datetime.timedelta(days=5, hours=2, minutes=15))
        expected_value = datetime.timedelta(days=10, hours=10, minutes=45)
        self.assertEqual(bind_vars["value"].getvalue(), expected_value)

    def test_1804_bind_in_out_fractional_second(self):
        "1804 - test binding in/out with set input sizes defined"
        bind_vars = \
                self.cursor.setinputsizes(value=oracledb.DB_TYPE_INTERVAL_DS)
        self.cursor.execute("""
                begin
                    :value := :value + to_dsinterval('5 08:30:00');
                end;""",
                value=datetime.timedelta(days=5, seconds=12.123789))
        expected_value = datetime.timedelta(days=10, hours=8, minutes=30,
                                            seconds=12, microseconds=123789)
        self.assertEqual(bind_vars["value"].getvalue(), expected_value)

    def test_1805_bind_out_var(self):
        "1805 - test binding out with cursor.var() method"
        var = self.cursor.var(oracledb.DB_TYPE_INTERVAL_DS)
        self.cursor.execute("""
                begin
                    :value := to_dsinterval('15 18:35:45.586');
                end;""",
                value=var)
        expected_value = datetime.timedelta(days=15, hours=18, minutes=35,
                                            seconds=45, milliseconds=586)
        self.assertEqual(var.getvalue(), expected_value)

    def test_1806_bind_in_out_var_direct_set(self):
        "1806 - test binding in/out with cursor.var() method"
        var = self.cursor.var(oracledb.DB_TYPE_INTERVAL_DS)
        var.setvalue(0, datetime.timedelta(days=1, minutes=50))
        self.cursor.execute("""
                begin
                    :value := :value + to_dsinterval('8 05:15:00');
                end;""",
                value=var)
        expected_value = datetime.timedelta(days=9, hours=6, minutes=5)
        self.assertEqual(var.getvalue(), expected_value)

    def test_1807_cursor_description(self):
        "1807 - test cursor description is accurate"
        self.cursor.execute("select * from TestIntervals")
        expected_value = [
            ('INTCOL', oracledb.DB_TYPE_NUMBER, 10, None, 9, 0, False),
            ('INTERVALCOL', oracledb.DB_TYPE_INTERVAL_DS, None, None, 2, 6,
                    False),
            ('NULLABLECOL', oracledb.DB_TYPE_INTERVAL_DS, None, None, 2, 6,
                    True)
        ]
        self.assertEqual(self.cursor.description, expected_value)

    def test_1808_fetchall(self):
        "1808 - test that fetching all of the data returns the correct results"
        self.cursor.execute("select * From TestIntervals order by IntCol")
        self.assertEqual(self.cursor.fetchall(), self.raw_data)
        self.assertEqual(self.cursor.fetchall(), [])

    def test_1809_fetchmany(self):
        "1809 - test that fetching data in chunks returns the correct results"
        self.cursor.execute("select * From TestIntervals order by IntCol")
        self.assertEqual(self.cursor.fetchmany(3), self.raw_data[0:3])
        self.assertEqual(self.cursor.fetchmany(2), self.raw_data[3:5])
        self.assertEqual(self.cursor.fetchmany(4), self.raw_data[5:9])
        self.assertEqual(self.cursor.fetchmany(3), self.raw_data[9:])
        self.assertEqual(self.cursor.fetchmany(3), [])

    def test_1810_fetchone(self):
        "1810 - test that fetching a single row returns the correct results"
        self.cursor.execute("""
                select *
                from TestIntervals
                where IntCol in (3, 4)
                order by IntCol""")
        self.assertEqual(self.cursor.fetchone(), self.data_by_key[3])
        self.assertEqual(self.cursor.fetchone(), self.data_by_key[4])
        self.assertEqual(self.cursor.fetchone(), None)

    def test_1811_bind_and_fetch_negative_interval(self):
        "1811 - test binding and fetching a negative interval"
        value = datetime.timedelta(days=-1, seconds=86314, microseconds=431152)
        self.cursor.execute("select :1 from dual", [value])
        result, = self.cursor.fetchone()
        self.assertEqual(result, value)

    def test_1812_unsupported_year_to_month_interval(self):
        "1812 - test year to month interval not supported"
        statement = "select INTERVAL '10-2' YEAR TO MONTH from dual"
        self.assertRaisesRegex(oracledb.NotSupportedError, "^DPY-3007:",
                               self.cursor.execute, statement)

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2021, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# data_types.pyx
#
# Cython file defining the data type array sent to the database server during
# connect (embedded in thin_impl.pyx).
#------------------------------------------------------------------------------

ctypedef struct DataType:
    uint16_t data_type
    uint16_t conv_data_type
    uint16_t representation

cdef DataType[312] DATA_TYPES = [
    [TNS_DATA_TYPE_VARCHAR, TNS_DATA_TYPE_VARCHAR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_NUMBER, TNS_DATA_TYPE_NUMBER, TNS_TYPE_REP_ORACLE],
    [TNS_DATA_TYPE_LONG, TNS_DATA_TYPE_LONG, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DATE, TNS_DATA_TYPE_DATE, TNS_TYPE_REP_ORACLE],
    [TNS_DATA_TYPE_RAW, TNS_DATA_TYPE_RAW, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_LONG_RAW, TNS_DATA_TYPE_LONG_RAW, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_UB2, TNS_DATA_TYPE_UB2, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_UB4, TNS_DATA_TYPE_UB4, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SB1, TNS_DATA_TYPE_SB1, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SB2, TNS_DATA_TYPE_SB2, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SB4, TNS_DATA_TYPE_SB4, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SWORD, TNS_DATA_TYPE_SWORD, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_UWORD, TNS_DATA_TYPE_UWORD, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PTRB, TNS_DATA_TYPE_PTRB, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PTRW, TNS_DATA_TYPE_PTRW, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_TIDDEF, TNS_DATA_TYPE_TIDDEF, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_ROWID, TNS_DATA_TYPE_ROWID, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AMS, TNS_DATA_TYPE_AMS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_BRN, TNS_DATA_TYPE_BRN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_CWD, TNS_DATA_TYPE_CWD, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_NEW_OAC, TNS_DATA_TYPE_NEW_OAC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OER8, TNS_DATA_TYPE_OER8, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_FUN, TNS_DATA_TYPE_FUN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AUA, TNS_DATA_TYPE_AUA, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RXH7, TNS_DATA_TYPE_RXH7, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_NA6, TNS_DATA_TYPE_NA6, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_BRP, TNS_DATA_TYPE_BRP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_BRV, TNS_DATA_TYPE_BRV, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KVA, TNS_DATA_TYPE_KVA, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_CLS, TNS_DATA_TYPE_CLS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_CUI, TNS_DATA_TYPE_CUI, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DFN, TNS_DATA_TYPE_DFN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DQR, TNS_DATA_TYPE_DQR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSC, TNS_DATA_TYPE_DSC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_EXE, TNS_DATA_TYPE_EXE, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_FCH, TNS_DATA_TYPE_FCH, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_GBV, TNS_DATA_TYPE_GBV, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_GEM, TNS_DATA_TYPE_GEM, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_GIV, TNS_DATA_TYPE_GIV, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OKG, TNS_DATA_TYPE_OKG, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_HMI, TNS_DATA_TYPE_HMI, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_INO, TNS_DATA_TYPE_INO, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_LNF, TNS_DATA_TYPE_LNF, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_ONT, TNS_DATA_TYPE_ONT, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OPE, TNS_DATA_TYPE_OPE, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OSQ, TNS_DATA_TYPE_OSQ, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SFE, TNS_DATA_TYPE_SFE, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SPF, TNS_DATA_TYPE_SPF, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_VSN, TNS_DATA_TYPE_VSN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_UD7, TNS_DATA_TYPE_UD7, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSA, TNS_DATA_TYPE_DSA, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PIN, TNS_DATA_TYPE_PIN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PFN, TNS_DATA_TYPE_PFN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PPT, TNS_DATA_TYPE_PPT, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_STO, TNS_DATA_TYPE_STO, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_ARC, TNS_DATA_TYPE_ARC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_MRS, TNS_DATA_TYPE_MRS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_MRT, TNS_DATA_TYPE_MRT, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_MRG, TNS_DATA_TYPE_MRG, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_MRR, TNS_DATA_TYPE_MRR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_MRC, TNS_DATA_TYPE_MRC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_VER, TNS_DATA_TYPE_VER, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_LON2, TNS_DATA_TYPE_LON2, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_INO2, TNS_DATA_TYPE_INO2, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_ALL, TNS_DATA_TYPE_ALL, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_UDB, TNS_DATA_TYPE_UDB, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AQI, TNS_DATA_TYPE_AQI, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_ULB, TNS_DATA_TYPE_ULB, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_ULD, TNS_DATA_TYPE_ULD, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SID, TNS_DATA_TYPE_SID, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_NA7, TNS_DATA_TYPE_NA7, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AL7, TNS_DATA_TYPE_AL7, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_K2RPC, TNS_DATA_TYPE_K2RPC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XDP, TNS_DATA_TYPE_XDP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OKO8, TNS_DATA_TYPE_OKO8, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_UD12, TNS_DATA_TYPE_UD12, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AL8, TNS_DATA_TYPE_AL8, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_LFOP, TNS_DATA_TYPE_LFOP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_FCRT, TNS_DATA_TYPE_FCRT, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DNY, TNS_DATA_TYPE_DNY, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OPR, TNS_DATA_TYPE_OPR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PLS, TNS_DATA_TYPE_PLS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XID, TNS_DATA_TYPE_XID, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_TXN, TNS_DATA_TYPE_TXN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DCB, TNS_DATA_TYPE_DCB, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_CCA, TNS_DATA_TYPE_CCA, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_WRN, TNS_DATA_TYPE_WRN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_TLH, TNS_DATA_TYPE_TLH, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_TOH, TNS_DATA_TYPE_TOH, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_FOI, TNS_DATA_TYPE_FOI, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SID2, TNS_DATA_TYPE_SID2, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_TCH, TNS_DATA_TYPE_TCH, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PII, TNS_DATA_TYPE_PII, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PFI, TNS_DATA_TYPE_PFI, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PPU, TNS_DATA_TYPE_PPU, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PTE, TNS_DATA_TYPE_PTE, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RXH8, TNS_DATA_TYPE_RXH8, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_N12, TNS_DATA_TYPE_N12, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AUTH, TNS_DATA_TYPE_AUTH, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KVAL, TNS_DATA_TYPE_KVAL, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_FGI, TNS_DATA_TYPE_FGI, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSY, TNS_DATA_TYPE_DSY, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYR8, TNS_DATA_TYPE_DSYR8, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYH8, TNS_DATA_TYPE_DSYH8, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYL, TNS_DATA_TYPE_DSYL, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYT8, TNS_DATA_TYPE_DSYT8, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYV8, TNS_DATA_TYPE_DSYV8, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYP, TNS_DATA_TYPE_DSYP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYF, TNS_DATA_TYPE_DSYF, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYK, TNS_DATA_TYPE_DSYK, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYY, TNS_DATA_TYPE_DSYY, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYQ, TNS_DATA_TYPE_DSYQ, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYC, TNS_DATA_TYPE_DSYC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYA, TNS_DATA_TYPE_DSYA, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OT8, TNS_DATA_TYPE_OT8, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYTY, TNS_DATA_TYPE_DSYTY, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AQE, TNS_DATA_TYPE_AQE, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KV, TNS_DATA_TYPE_KV, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AQD, TNS_DATA_TYPE_AQD, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AQ8, TNS_DATA_TYPE_AQ8, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RFS, TNS_DATA_TYPE_RFS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RXH10, TNS_DATA_TYPE_RXH10, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPN, TNS_DATA_TYPE_KPN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPDNR, TNS_DATA_TYPE_KPDNR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYD, TNS_DATA_TYPE_DSYD, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYS, TNS_DATA_TYPE_DSYS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYR, TNS_DATA_TYPE_DSYR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYH, TNS_DATA_TYPE_DSYH, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYT, TNS_DATA_TYPE_DSYT, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYV, TNS_DATA_TYPE_DSYV, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AQM, TNS_DATA_TYPE_AQM, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OER11, TNS_DATA_TYPE_OER11, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AQL, TNS_DATA_TYPE_AQL, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OTC, TNS_DATA_TYPE_OTC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KFNO, TNS_DATA_TYPE_KFNO, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KFNP, TNS_DATA_TYPE_KFNP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KGT8, TNS_DATA_TYPE_KGT8, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RASB4, TNS_DATA_TYPE_RASB4, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RAUB2, TNS_DATA_TYPE_RAUB2, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RAUB1, TNS_DATA_TYPE_RAUB1, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RATXT, TNS_DATA_TYPE_RATXT, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RSSB4, TNS_DATA_TYPE_RSSB4, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RSUB2, TNS_DATA_TYPE_RSUB2, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RSUB1, TNS_DATA_TYPE_RSUB1, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RSTXT, TNS_DATA_TYPE_RSTXT, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RIDL, TNS_DATA_TYPE_RIDL, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_GLRDD, TNS_DATA_TYPE_GLRDD, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_GLRDG, TNS_DATA_TYPE_GLRDG, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_GLRDC, TNS_DATA_TYPE_GLRDC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OKO, TNS_DATA_TYPE_OKO, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DPP, TNS_DATA_TYPE_DPP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DPLS, TNS_DATA_TYPE_DPLS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DPMOP, TNS_DATA_TYPE_DPMOP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_STAT, TNS_DATA_TYPE_STAT, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RFX, TNS_DATA_TYPE_RFX, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_FAL, TNS_DATA_TYPE_FAL, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_CKV, TNS_DATA_TYPE_CKV, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DRCX, TNS_DATA_TYPE_DRCX, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KGH, TNS_DATA_TYPE_KGH, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AQO, TNS_DATA_TYPE_AQO, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OKGT, TNS_DATA_TYPE_OKGT, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPFC, TNS_DATA_TYPE_KPFC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_FE2, TNS_DATA_TYPE_FE2, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SPFP, TNS_DATA_TYPE_SPFP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DPULS, TNS_DATA_TYPE_DPULS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AQA, TNS_DATA_TYPE_AQA, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPBF, TNS_DATA_TYPE_KPBF, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_TSM, TNS_DATA_TYPE_TSM, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_MSS, TNS_DATA_TYPE_MSS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPC, TNS_DATA_TYPE_KPC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_CRS, TNS_DATA_TYPE_CRS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KKS, TNS_DATA_TYPE_KKS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KSP, TNS_DATA_TYPE_KSP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KSPTOP, TNS_DATA_TYPE_KSPTOP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KSPVAL, TNS_DATA_TYPE_KSPVAL, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PSS, TNS_DATA_TYPE_PSS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_NLS, TNS_DATA_TYPE_NLS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_ALS, TNS_DATA_TYPE_ALS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KSDEVTVAL, TNS_DATA_TYPE_KSDEVTVAL, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KSDEVTTOP, TNS_DATA_TYPE_KSDEVTTOP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPSPP, TNS_DATA_TYPE_KPSPP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KOL, TNS_DATA_TYPE_KOL, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_LST, TNS_DATA_TYPE_LST, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_ACX, TNS_DATA_TYPE_ACX, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SCS, TNS_DATA_TYPE_SCS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RXH, TNS_DATA_TYPE_RXH, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPDNS, TNS_DATA_TYPE_KPDNS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPDCN, TNS_DATA_TYPE_KPDCN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPNNS, TNS_DATA_TYPE_KPNNS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPNCN, TNS_DATA_TYPE_KPNCN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPS, TNS_DATA_TYPE_KPS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_APINF, TNS_DATA_TYPE_APINF, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_TEN, TNS_DATA_TYPE_TEN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSSCS, TNS_DATA_TYPE_XSSCS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSSSO, TNS_DATA_TYPE_XSSSO, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSSAO, TNS_DATA_TYPE_XSSAO, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KSRPC, TNS_DATA_TYPE_KSRPC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KVL, TNS_DATA_TYPE_KVL, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSSDEF, TNS_DATA_TYPE_XSSDEF, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PDQCINV, TNS_DATA_TYPE_PDQCINV, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PDQIDC, TNS_DATA_TYPE_PDQIDC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPDQCSTA, TNS_DATA_TYPE_KPDQCSTA, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPRS, TNS_DATA_TYPE_KPRS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPDQIDC, TNS_DATA_TYPE_KPDQIDC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RTSTRM, TNS_DATA_TYPE_RTSTRM, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SESSGET, TNS_DATA_TYPE_SESSGET, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SESSREL, TNS_DATA_TYPE_SESSREL, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SESSRET, TNS_DATA_TYPE_SESSRET, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SCN6, TNS_DATA_TYPE_SCN6, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KECPA, TNS_DATA_TYPE_KECPA, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KECPP, TNS_DATA_TYPE_KECPP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SXA, TNS_DATA_TYPE_SXA, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KVARR, TNS_DATA_TYPE_KVARR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPNGN, TNS_DATA_TYPE_KPNGN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_BINARY_INTEGER, TNS_DATA_TYPE_NUMBER, TNS_TYPE_REP_ORACLE],
    [TNS_DATA_TYPE_FLOAT, TNS_DATA_TYPE_NUMBER, TNS_TYPE_REP_ORACLE],
    [TNS_DATA_TYPE_STR, TNS_DATA_TYPE_VARCHAR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_VNU, TNS_DATA_TYPE_NUMBER, TNS_TYPE_REP_ORACLE],
    [TNS_DATA_TYPE_PDN, TNS_DATA_TYPE_NUMBER, TNS_TYPE_REP_ORACLE],
    [TNS_DATA_TYPE_VCS, TNS_DATA_TYPE_VARCHAR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_VBI, TNS_DATA_TYPE_VARCHAR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OAC, TNS_DATA_TYPE_NEW_OAC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_UIN, TNS_DATA_TYPE_NUMBER, TNS_TYPE_REP_ORACLE],
    [TNS_DATA_TYPE_SLS, TNS_DATA_TYPE_NUMBER, TNS_TYPE_REP_ORACLE],
    [TNS_DATA_TYPE_LVC, TNS_DATA_TYPE_VARCHAR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_LVB, TNS_DATA_TYPE_RAW, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_CHAR, TNS_DATA_TYPE_CHAR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AVC, TNS_DATA_TYPE_CHAR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_BINARY_FLOAT, TNS_DATA_TYPE_BINARY_FLOAT,
            TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_BINARY_DOUBLE, TNS_DATA_TYPE_BINARY_DOUBLE,
            TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_CURSOR, TNS_DATA_TYPE_CURSOR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RDD, TNS_DATA_TYPE_ROWID, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OSL, TNS_DATA_TYPE_OSL, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_EXT_NAMED, TNS_DATA_TYPE_INT_NAMED, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_INT_NAMED, TNS_DATA_TYPE_INT_NAMED, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_EXT_REF, TNS_DATA_TYPE_INT_REF, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_INT_REF, TNS_DATA_TYPE_INT_REF, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_CLOB, TNS_DATA_TYPE_CLOB, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_BLOB, TNS_DATA_TYPE_BLOB, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_BFILE, TNS_DATA_TYPE_BFILE, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_CFILE, TNS_DATA_TYPE_CFILE, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RSET, TNS_DATA_TYPE_CURSOR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_JSON, TNS_DATA_TYPE_JSON, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DJSON, TNS_DATA_TYPE_DJSON, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_CLV, TNS_DATA_TYPE_CLV, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DTR, TNS_DATA_TYPE_NUMBER, TNS_TYPE_REP_ORACLE],
    [TNS_DATA_TYPE_DUN, TNS_DATA_TYPE_NUMBER, TNS_TYPE_REP_ORACLE],
    [TNS_DATA_TYPE_DOP, TNS_DATA_TYPE_NUMBER, TNS_TYPE_REP_ORACLE],
    [TNS_DATA_TYPE_VST, TNS_DATA_TYPE_VARCHAR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_ODT, TNS_DATA_TYPE_DATE, TNS_TYPE_REP_ORACLE],
    [TNS_DATA_TYPE_DOL, TNS_DATA_TYPE_NUMBER, TNS_TYPE_REP_ORACLE],
    [TNS_DATA_TYPE_TIME, TNS_DATA_TYPE_TIME, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_TIME_TZ, TNS_DATA_TYPE_TIME_TZ, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_TIMESTAMP, TNS_DATA_TYPE_TIMESTAMP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_TIMESTAMP_TZ, TNS_DATA_TYPE_TIMESTAMP_TZ,
            TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_INTERVAL_YM, TNS_DATA_TYPE_INTERVAL_YM,
            TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_INTERVAL_DS, TNS_DATA_TYPE_INTERVAL_DS,
            TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_EDATE, TNS_DATA_TYPE_DATE, TNS_TYPE_REP_ORACLE],
    [TNS_DATA_TYPE_ETIME, TNS_DATA_TYPE_ETIME, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_ETTZ, TNS_DATA_TYPE_ETTZ, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_ESTAMP, TNS_DATA_TYPE_ESTAMP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_ESTZ, TNS_DATA_TYPE_ESTZ, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_EIYM, TNS_DATA_TYPE_EIYM, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_EIDS, TNS_DATA_TYPE_EIDS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DCLOB, TNS_DATA_TYPE_CLOB, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DBLOB, TNS_DATA_TYPE_BLOB, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DBFILE, TNS_DATA_TYPE_BFILE, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_UROWID, TNS_DATA_TYPE_UROWID, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_TIMESTAMP_LTZ, TNS_DATA_TYPE_TIMESTAMP_LTZ,
            TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_ESITZ, TNS_DATA_TYPE_TIMESTAMP_LTZ, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_UB8, TNS_DATA_TYPE_UB8, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PNTY, TNS_DATA_TYPE_INT_NAMED, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_BOOLEAN, TNS_DATA_TYPE_BOOLEAN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSNSOP, TNS_DATA_TYPE_XSNSOP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSATTR, TNS_DATA_TYPE_XSATTR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSNS, TNS_DATA_TYPE_XSNS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_UB1ARRAY, TNS_DATA_TYPE_UB1ARRAY, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SESSSTATE, TNS_DATA_TYPE_SESSSTATE, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AC_REPLAY, TNS_DATA_TYPE_AC_REPLAY, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AC_CONT, TNS_DATA_TYPE_AC_CONT, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_IMPLRES, TNS_DATA_TYPE_IMPLRES, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OER, TNS_DATA_TYPE_OER, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_TXT, TNS_DATA_TYPE_TXT, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSSESSNS, TNS_DATA_TYPE_XSSESSNS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSATTOP, TNS_DATA_TYPE_XSATTOP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSCREOP, TNS_DATA_TYPE_XSCREOP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSDETOP, TNS_DATA_TYPE_XSDETOP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSDESOP, TNS_DATA_TYPE_XSDESOP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSSETSP, TNS_DATA_TYPE_XSSETSP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSSIDP, TNS_DATA_TYPE_XSSIDP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSPRIN, TNS_DATA_TYPE_XSPRIN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSKVL, TNS_DATA_TYPE_XSKVL, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSSSDEF2, TNS_DATA_TYPE_XSSSDEF2, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSNSOP2, TNS_DATA_TYPE_XSNSOP2, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSNS2, TNS_DATA_TYPE_XSNS2, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPDNREQ, TNS_DATA_TYPE_KPDNREQ, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPDNRNF, TNS_DATA_TYPE_KPDNRNF, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPNGNC, TNS_DATA_TYPE_KPNGNC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPNRI, TNS_DATA_TYPE_KPNRI, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AQENQ, TNS_DATA_TYPE_AQENQ, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AQDEQ, TNS_DATA_TYPE_AQDEQ, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AQJMS, TNS_DATA_TYPE_AQJMS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPDNRPAY, TNS_DATA_TYPE_KPDNRPAY, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPDNRACK, TNS_DATA_TYPE_KPDNRACK, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPDNRMP, TNS_DATA_TYPE_KPDNRMP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPDNRDQ, TNS_DATA_TYPE_KPDNRDQ, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SCN, TNS_DATA_TYPE_SCN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SCN8, TNS_DATA_TYPE_SCN8, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_CHUNKINFO, TNS_DATA_TYPE_CHUNKINFO, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_UDS, TNS_DATA_TYPE_UDS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_TNP, TNS_DATA_TYPE_TNP, TNS_TYPE_REP_UNIVERSAL],
    [0, 0, 0]
]
